@@ -1,5 +1,7 @@
 ;;; vm-menu.el --- Menu related functions and commands
 ;;
+;; This file is part of VM
+;;
 ;; Copyright (C) 1994 Heiko Muenkel
 ;; Copyright (C) 1995, 1997 Kyle E. Jones
 ;; Copyright (C) 2003-2006 Robert Widhopf-Fenk
@@ -48,15 +50,56 @@
 ;;    Removed the need for -A in ls flags.
 ;;    Some systems' ls don't support -A.
 
-(require 'vm-easymenu)
-
 ;;; Code:
+
+(provide 'vm-menu)
+
+(eval-when-compile
+  (require 'vm-misc)
+  (require 'vm-mime)
+
+  (defvar current-menubar nil))
+
+(declare-function event-window "vm-xemacs" (event))
+(declare-function event-point "vm-xemacs" (event))
+(declare-function popup-mode-menu "vm-xemacs" (&optional event))
+(declare-function event-closest-point "vm-xemacs" (event))
+(declare-function find-menu-item "vm-xemacs" 
+		  (menubar item-path-list &optional parent))
+(declare-function add-menu-button "vm-xemacs" 
+		  (menu-path menu-leaf &optional before in-menu))
+(declare-function add-menu-item "vm-xemacs" 
+		  (menu-path item-name function enabled-p &optional before))
+(declare-function add-menu "vm-xemacs" 
+		  (menu-path menu-name menu-items &optional before))
+(declare-function set-menubar-dirty-flag "vm-xemacs" ())
+(declare-function set-buffer-menubar "vm-xemacs" (menubar))
+
+
+(declare-function vm-pop-find-name-for-spec "vm-pop" (spec))
+(declare-function vm-imap-folder-for-spec "vm-imap" (spec))
+(declare-function vm-mime-plain-message-p "vm-mime" (message))
+(declare-function vm-yank-message "vm-reply" (message))
+(declare-function vm-mail "vm" (&optional to subject))
+(declare-function vm-get-header-contents "vm-summary"
+		  (message header-name-regexp &optional clump-sep))
+(declare-function vm-mail-mode-get-header-contents "vm-reply"
+		  (header-name-regexp))
+(declare-function vm-create-virtual-folder "vm-virtual"
+		  (selector &optional arg read-only name bookmark))
+(declare-function vm-create-virtual-folder-of-threads "vm-virtual"
+		  (selector &optional arg read-only name bookmark))
+(declare-function vm-so-sortable-subject "vm-sort" (message))
+(declare-function vm-su-from "vm-summary" (message))
+
+
+;; This will be extended in code
 (defvar vm-menu-folders-menu
   '("Manipulate Folders"
     ["Make Folders Menu" vm-menu-hm-make-folder-menu vm-folder-directory])
   "VM folder menu list.")
 
-(defvar vm-menu-folder-menu
+(defconst vm-menu-folder-menu
   `("Folder"
     ,(if vm-fsfemacs-p
 	["Manipulate Folders" ignore (ignore)]
@@ -73,12 +116,12 @@
     ["Expunge" vm-expunge-folder vm-message-list]
     ["Expunge POP Messages" vm-expunge-pop-messages
      (vm-menu-can-expunge-pop-messages-p)]
-    ["Expunge IMAP Messages" vm-expunge-pop-messages
+    ["Expunge IMAP Messages" vm-expunge-imap-messages
      (vm-menu-can-expunge-imap-messages-p)]
     "---"
     ["Visit Local Folder" vm-visit-folder t]
     ["Visit POP Folder" vm-visit-pop-folder vm-pop-folder-alist]
-    ["Visit IMAP Folder" vm-visit-imap-folder vm-imap-server-list]
+    ["Visit IMAP Folder" vm-visit-imap-folder vm-imap-account-alist]
     ["Revert Folder (back to disk version)" vm-revert-buffer
      (vm-menu-can-revert-p)]
     ["Recover Folder (from auto-save file)" vm-recover-file
@@ -88,13 +131,13 @@
     ["Quit" vm-quit-no-change t]
     ["Save & Quit" vm-quit t]
     "---"
-    "---"
+    ;; "---"
     ;; special string that marks the tail of this menu for
     ;; vm-menu-install-visited-folders-menu.
     "-------"
     ))
 
-(defvar vm-menu-dispose-menu
+(defconst vm-menu-dispose-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Dispose"
 			 "Dispose"
@@ -109,6 +152,7 @@
       ["Reply to All (citing original)" vm-followup-include-text
        vm-message-list]
       ["Forward" vm-forward-message vm-message-list]
+      ["Forward in Plain Text" vm-forward-message-plain vm-message-list]
       ["Resend" vm-resend-message vm-message-list]
       ["Retry Bounce" vm-resend-bounced-message vm-message-list]
       "---"
@@ -116,16 +160,18 @@
       ["Delete" vm-delete-message vm-message-list]
       ["Undelete"	vm-undelete-message vm-message-list]
       ["Kill Current Subject" vm-kill-subject vm-message-list]
-      ["Mark Unread" vm-unread-message vm-message-list]
+      ["Mark Unread" vm-mark-message-unread vm-message-list]
       ["Edit" vm-edit-message vm-message-list]
       ["Print" vm-print-message vm-message-list]
       ["Pipe to Command" vm-pipe-message-to-command vm-message-list]
+      ["Attach to Message Composition"
+       vm-attach-message-to-composition vm-message-list] 
       "---"
       ["Burst Message as Digest" (vm-burst-digest "guess") vm-message-list]
       ["Decode MIME" vm-decode-mime-message (vm-menu-can-decode-mime-p)]
       )))
 
-(defvar vm-menu-motion-menu
+(defconst vm-menu-motion-menu
   '("Motion"
     ["Page Up" vm-scroll-backward vm-message-list]
     ["Page Down" vm-scroll-forward vm-message-list]
@@ -135,7 +181,6 @@
     "----"
     ["Expose/Hide Headers" vm-expose-hidden-headers vm-message-list]
     "----"
-    "----"
     ["Next Message" vm-next-message t]
     ["Previous Message"	vm-previous-message t]
     "---"
@@ -153,21 +198,38 @@
     ["Go to Parent Message" vm-goto-parent-message t]
     ))
 
-(defvar vm-menu-virtual-menu
+(defconst vm-menu-virtual-menu
   '("Virtual"
     ["Visit Virtual Folder" vm-visit-virtual-folder t]
-    ["Visit Virtual Folder Same Author" vm-visit-virtual-folder-same-author t]
-    ["Visit Virtual Folder Same Subject" vm-visit-virtual-folder-same-subject t]
-    ["Create Virtual Folder" vm-create-virtual-folder t]
-    ["Apply Virtual Folder" vm-apply-virtual-folder t]
+    ["Apply Virtual Folder Selectors" vm-apply-virtual-folder t]
+    ["Omit Message" vm-virtual-omit-message t]
+    ["Update all" vm-virtual-update-folders]
     "---"
+    "Search Folders"
+    ["Author" vm-create-author-virtual-folder t]
+    ["Recipients" vm-create-author-or-recipient-virtual-folder t]
+    ["Subject" vm-create-subject-virtual-folder t]
+    ["Text (Body)" vm-create-text-virtual-folder t]
+    ["Days" vm-create-date-virtual-folder t]
+    ["Label" vm-create-label-virtual-folder t]
+    ["Flagged" vm-create-flagged-virtual-folder t]
+    ["Unseen" vm-create-unseen-virtual-folder t]
+    ["Same Author as current" vm-create-virtual-folder-same-author t]
+    ["Same Subject as current" vm-create-virtual-folder-same-subject t]
+    ["Create General" vm-create-virtual-folder t]
+    ["Create General (Threads)" vm-create-virtual-folder-of-threads t]
     "---"
+    "Auto operations"
+    ["Delete Message(s)" vm-virtual-auto-delete-message t]
+    ["Save Message(s)" vm-virtual-save-message t]
+    ["Archive Messages" vm-virtual-auto-archive-messages t]
+    
     ;; special string that marks the tail of this menu for
     ;; vm-menu-install-known-virtual-folders-menu.
     "-------"
     ))
 
-(defvar vm-menu-send-menu
+(defconst vm-menu-send-menu
   '("Send"
     ["Compose" vm-mail t]
     ["Continue Composing" vm-continue-composing-message vm-message-list]
@@ -176,6 +238,7 @@
     ["Reply to Author (citing original)" vm-reply-include-text vm-message-list]
     ["Reply to All (citing original)" vm-followup-include-text vm-message-list]
     ["Forward Message" vm-forward-message vm-message-list]
+    ["Forward Message in Plain Text" vm-forward-message-plain vm-message-list]
     ["Resend Message" vm-resend-message vm-message-list]
     ["Retry Bounced Message" vm-resend-bounced-message vm-message-list]
     ["Send Digest (RFC934)" vm-send-rfc934-digest vm-message-list]
@@ -183,7 +246,7 @@
     ["Send MIME Digest" vm-send-mime-digest vm-message-list]
     ))
 
-(defvar vm-menu-mark-menu
+(defconst vm-menu-mark-menu
   '("Mark"
     ["Next Command Uses Marks..." vm-next-command-uses-marks
      :active vm-message-list
@@ -207,40 +270,55 @@
     ["Unmark Thread Subtree" vm-unmark-thread-subtree vm-message-list]
     ))
 
-(defvar vm-menu-label-menu
+(defconst vm-menu-label-menu
   '("Label"
     ["Add Label" vm-add-message-labels vm-message-list]
     ["Add Existing Label" vm-add-existing-message-labels vm-message-list]
     ["Remove Label" vm-delete-message-labels vm-message-list]
     ))
 
-(defvar vm-menu-sort-menu
+(defconst vm-menu-sort-menu
   '("Sort"
+    "By ascending"
+    "---"
+    ["Date" (vm-sort-messages "date") vm-message-list]
+    ["Activity" (vm-sort-messages "activity") vm-message-list]
+    ["Subject" (vm-sort-messages "subject") vm-message-list]
+    ["Author" (vm-sort-messages "author") vm-message-list]
+    ["Recipients" (vm-sort-messages "recipients") vm-message-list]
+    ["Lines" (vm-sort-messages "line-count") vm-message-list]
+    ["Bytes" (vm-sort-messages "byte-count") vm-message-list]
+    "---"
+    "By descending"
+    "---"
+    ["Date" (vm-sort-messages "reversed-date") vm-message-list]
+    ["Activity" (vm-sort-messages "reversed-activity") vm-message-list]
+    ["Subject" (vm-sort-messages "reversed-subject") vm-message-list]
+    ["Author" (vm-sort-messages "reversed-author") vm-message-list]
+    ["Recipients" (vm-sort-messages "reversed-recipients") vm-message-list]
+    ["Lines" (vm-sort-messages "reversed-line-count") vm-message-list]
+    ["Bytes" (vm-sort-messages "reversed-byte-count") vm-message-list]
+    "---"
     ["By Multiple Fields..." vm-sort-messages vm-message-list]
-    "---"
-    ["By Date" (vm-sort-messages "date") vm-message-list]
-    ["By Subject" (vm-sort-messages "subject") vm-message-list]
-    ["By Author" (vm-sort-messages "author") vm-message-list]
-    ["By Recipients" (vm-sort-messages "recipients") vm-message-list]
-    ["By Lines" (vm-sort-messages "line-count") vm-message-list]
-    ["By Bytes" (vm-sort-messages "byte-count") vm-message-list]
-    "---"
-    ["By Date (backward)" (vm-sort-messages "reversed-date") vm-message-list]
-    ["By Subject (backward)" (vm-sort-messages "reversed-subject") vm-message-list]
-    ["By Author (backward)" (vm-sort-messages "reversed-author") vm-message-list]
-    ["By Recipients (backward)" (vm-sort-messages "reversed-recipients") vm-message-list]
-    ["By Lines (backward)" (vm-sort-messages "reversed-line-count") vm-message-list]
-    ["By Bytes (backward)" (vm-sort-messages "reversed-byte-count") vm-message-list]
+    ["Revert to Physical Order" (vm-sort-messages "physical-order" t) vm-message-list]
     "---"
     ["Toggle Threading" vm-toggle-threads-display t]
-    "---"
-    ["Revert to Physical Order" (vm-sort-messages "physical-order" t) vm-message-list]
+    ["Expand/Collapse Thread" vm-toggle-thread t]
+    ["Expand All Threads" vm-expand-all-threads t]
+    ["Collapse All Threads" vm-collapse-all-threads t]
     ))
 
-(defvar vm-menu-help-menu
-  '("Help!"
+(defconst vm-menu-help-menu
+  '("Help"
+    ["Switch to Emacs Menubar" vm-menu-toggle-menubar t]
+    "---"
+    ["Customize VM" vm-customize t]
+    ["Describe VM Mode" describe-mode t]
+    ["VM News" vm-view-news t]
+    ["VM Manual" vm-menu-view-manual t]
+    ["Submit Bug Report" vm-submit-bug-report t]
+    "---"
     ["What Now?" vm-help t]
-    ["Describe Mode" describe-mode t]
     ["Revert Folder (back to disk version)" revert-buffer (vm-menu-can-revert-p)]
     ["Recover Folder (from auto-save file)" recover-file (vm-menu-can-recover-p)]
     "---"
@@ -248,19 +326,32 @@
     ["Quit Without Saving" vm-quit-no-change t]
     ))
 
-(defvar vm-menu-undo-menu
-  ["Undo" vm-undo (vm-menu-can-undo-p)]
+(defconst vm-menu-xemacs-undo-button
+  ["[Undo]" vm-undo (vm-menu-can-undo-p)]
   )
 
-(defvar vm-menu-emacs-button
-  ["XEmacs" vm-menu-toggle-menubar t]
+(defconst vm-menu-undo-menu
+  '("Undo"
+    ["Undo" vm-undo (vm-menu-can-undo-p)]
+    )
+  "Undo menu for FSF Emacs builds that do not allow menubar buttons.")
+
+(defconst vm-menu-emacs-button
+  ["[Emacs Menubar]" vm-menu-toggle-menubar t]
   )
 
-(defvar vm-menu-vm-button
-  ["VM" vm-menu-toggle-menubar t]
+(defconst vm-menu-emacs-menu
+  '("Menubar"
+    ["Switch to Emacs Menubar" vm-menu-toggle-menubar t]
+    )
+  "Menu with a \"Swich to Emacs\" action meant for FSF Emacs builds that
+do not allow menubar buttons.")
+
+(defconst vm-menu-vm-button
+  ["[VM Menubar]" vm-menu-toggle-menubar t]
   )
 
-(defvar vm-menu-mail-menu
+(defconst vm-menu-mail-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Mail Commands"
 			 "Mail Commands"
@@ -367,17 +458,16 @@
 	:style radio
 	:selected (eq vm-mime-8bit-text-transfer-encoding 'base64)])
       "----"
-      ["Attach File..."	vm-mime-attach-file vm-send-using-mime]
-;;	   ["Attach MIME Message..." vm-mime-attach-mime-file
-;;	    vm-send-using-mime]
+      ["Attach File..."	vm-attach-file vm-send-using-mime]
+      ["Attach MIME Message..." vm-attach-mime-file vm-send-using-mime]
       ["Encode MIME, But Don't Send" vm-mime-encode-composition
        (and vm-send-using-mime
 	    (null (vm-mail-mode-get-header-contents "MIME-Version:")))]
-      ["Preview MIME Before Sending" vm-mime-preview-composition
+      ["Preview MIME Before Sending" vm-preview-composition
        vm-send-using-mime]
       )))
 
-(defvar vm-menu-mime-dispose-menu
+(defconst vm-menu-mime-dispose-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Take Action on MIME body ..."
 			 "Take Action on MIME body ..."
@@ -386,50 +476,52 @@
 		 (list "Take Action on MIME body ..."))))
     `(,@title
       ["Display as Text (in default face)"
-       (vm-mime-run-display-function-at-point
-	'vm-mime-display-body-as-text) t]
+       vm-mime-reader-map-display-using-default t]
       ["Display using External Viewer"
-       (vm-mime-run-display-function-at-point
-	'vm-mime-display-body-using-external-viewer) t]
+       vm-mime-reader-map-display-using-external-viewer t]
+      ["Convert to Text and Display"
+       vm-mime-reader-map-convert-then-display
+       (vm-menu-can-convert-to-text/plain (vm-mime-get-button-layout))]
       ;; FSF Emacs does not allow a non-string menu element name.
-      ,@(if (vm-menu-can-eval-item-name)
-	    (list [(format "Convert to %s and Display"
-			   (or (nth 1 (vm-mime-can-convert
-				       (car
-					(vm-mm-layout-type
-					 (vm-mime-get-button-layout e)))))
-			       "different type"))
-		   (vm-mime-run-display-function-at-point
-		    'vm-mime-convert-body-then-display)
-		   (vm-mime-can-convert
-		    (car (vm-mm-layout-type
-			  (vm-mime-get-button-layout e))))]))
+      ;; This is not working on XEmacs either.  USR, 2011-03-05
+      ;; ,@(if (vm-menu-can-eval-item-name)
+      ;; 	    (list [(format "Convert to %s and Display"
+      ;; 			   (or (nth 1 (vm-mime-can-convert
+      ;; 				       (car
+      ;; 					(vm-mm-layout-type
+      ;; 					 (vm-mime-get-button-layout)))))
+      ;; 			       "different type"))
+      ;; 		   (vm-mime-run-display-function-at-point
+      ;; 		    'vm-mime-convert-body-then-display)
+      ;; 		   (vm-mime-can-convert
+      ;; 		    (car (vm-mm-layout-type
+      ;; 			  (vm-mime-get-button-layout))))]))
       "---"
-      ["Save to File" vm-mime-reader-map-save-file t]
-      ["Save to Folder" vm-mime-reader-map-save-message
-       (let ((layout (vm-mime-run-display-function-at-point
-		      (function
-		       (lambda (e)
-			 (vm-extent-property e 'vm-mime-layout))))))
+      ["Undo" 
+       vm-undo]
+      "---"
+      ["Save to File" 
+       vm-mime-reader-map-save-file t]
+      ["Save to Folder" 
+       vm-mime-reader-map-save-message
+       (let ((layout (vm-mime-get-button-layout)))
 	 (if (null layout)
 	     nil
 	   (or (vm-mime-types-match "message/rfc822"
 				    (car (vm-mm-layout-type layout)))
 	       (vm-mime-types-match "message/news"
 				    (car (vm-mm-layout-type layout))))))]
-      ["Send to Printer" (vm-mime-run-display-function-at-point
-			  'vm-mime-send-body-to-printer) t]
+      ["Send to Printer" 
+       vm-mime-reader-map-pipe-to-printer t]
       ["Pipe to Shell Command (display output)"
-       (vm-mime-run-display-function-at-point
-	'vm-mime-pipe-body-to-queried-command) t]
+       vm-mime-reader-map-pipe-to-command t]
       ["Pipe to Shell Command (discard output)"
-       (vm-mime-run-display-function-at-point
-	'vm-mime-pipe-body-to-queried-command-discard-output) t]
+       vm-mime-reader-map-pipe-to-command-discard-output t]
       ["Attach to Message Composition Buffer"
-       vm-mime-attach-object-from-message t]
+       vm-mime-reader-map-attach-to-composition t]
       ["Delete" vm-delete-mime-object t])))
 
-(defvar vm-menu-url-browser-menu
+(defconst vm-menu-url-browser-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Send URL to ..."
 			 "Send URL to ..."
@@ -442,30 +534,52 @@
 		   'w3-fetch)
 		  (t 'w3-fetch-other-frame))))
     `(,@title
-      ["Emacs W3" (vm-mouse-send-url-at-position (point) (quote ,w3))
-       (fboundp (quote ,w3))]
-      ["Mosaic"
-       (vm-mouse-send-url-at-position (point)
-				      'vm-mouse-send-url-to-mosaic)
-       t]
-      ["mMosaic"
-       (vm-mouse-send-url-at-position (point)
-				      'vm-mouse-send-url-to-mmosaic)
-       t]
-      ["Netscape"
-       (vm-mouse-send-url-at-position (point)
-				      'vm-mouse-send-url-to-netscape)
-       t]
-      ["Konqueror"
-       (vm-mouse-send-url-at-position (point)
-				      'vm-mouse-send-url-to-konqueror)
+      ["Window system (Copy)"
+       (vm-mouse-send-url-at-position 
+	(point) 'vm-mouse-send-url-to-window-system)
        t]
       ["X Clipboard"
-       (vm-mouse-send-url-at-position (point)
-				      'vm-mouse-send-url-to-clipboard)
-       t])))
+       (vm-mouse-send-url-at-position 
+	(point) 'vm-mouse-send-url-to-clipboard)
+       t]
+      ["browse-url"
+       (vm-mouse-send-url-at-position (point) 'browse-url) 
+       browse-url-browser-function]
+      ["Emacs W3" (vm-mouse-send-url-at-position (point) (quote ,w3))
+       (fboundp (quote ,w3))]
+      ["Emacs W3M" (vm-mouse-send-url-at-position (point) 'w3m-browse-url)
+       (fboundp 'w3m-browse-url)]
+      "---"
+      ["Firefox"
+       (vm-mouse-send-url-at-position 
+	(point) 'vm-mouse-send-url-to-firefox)
+       vm-firefox-client-program]
+      ["Konqueror"
+       (vm-mouse-send-url-at-position 
+	(point) 'vm-mouse-send-url-to-konqueror)
+       vm-konqueror-client-program]
+      ;; ["Mosaic"
+      ;;  (vm-mouse-send-url-at-position 
+      ;;  (point) 'vm-mouse-send-url-to-mosaic)
+      ;;  vm-mosaic-program]
+      ;; ["mMosaic"
+      ;;  (vm-mouse-send-url-at-position 
+      ;; 	(point) 'vm-mouse-send-url-to-mmosaic)
+      ;;  vm-mmosaic-program]
+      ["Mozilla"
+       (vm-mouse-send-url-at-position 
+	(point) 'vm-mouse-send-url-to-mozilla)
+       vm-mozilla-program]
+;;       ["Netscape"
+;;        (vm-mouse-send-url-at-position 
+;; 	(point) 'vm-mouse-send-url-to-netscape)
+;;        vm-netscape-program]
+      ["Opera"
+       (vm-mouse-send-url-at-position 
+	(point) 'vm-mouse-send-url-to-opera)
+       vm-opera-program])))
 
-(defvar vm-menu-mailto-url-browser-menu
+(defconst vm-menu-mailto-url-browser-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Send Mail using ..."
 			 "Send Mail using ..."
@@ -475,7 +589,7 @@
     `(,@title
       ["VM" (vm-mouse-send-url-at-position (point) 'ignore) t])))
 
-(defvar vm-menu-subject-menu
+(defconst vm-menu-subject-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Take Action on Subject..."
 			 "Take Action on Subject..."
@@ -496,7 +610,7 @@
        vm-message-list]
       )))
 
-(defvar vm-menu-author-menu
+(defconst vm-menu-author-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Take Action on Author..."
 			 "Take Action on Author..."
@@ -510,9 +624,11 @@
        vm-message-list]
       ["Virtual Folder, Matching Author" vm-menu-create-author-virtual-folder
        vm-message-list]
+      ["Send a message" vm-menu-mail-to
+       vm-message-list]
       )))
 
-(defvar vm-menu-attachment-menu
+(defconst vm-menu-attachment-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Fiddle With Attachment"
 			 "Fiddle With Attachment"
@@ -602,7 +718,7 @@
        :style button]
       )))
 
-(defvar vm-menu-image-menu
+(defconst vm-menu-image-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "Redisplay Image"
 			 "Redisplay Image"
@@ -644,7 +760,7 @@
 
 (defvar vm-menu-vm-menubar nil)
 
-(defvar vm-menu-vm-menu
+(defconst vm-menu-vm-menu
   (let ((title (if (vm-menu-fsfemacs19-menus-p)
 		   (list "VM"
 			 "VM"
@@ -665,7 +781,9 @@
       "---"
       ,vm-menu-help-menu)))
 
-(defvar vm-mode-menu-map nil)
+(defvar vm-mode-menu-map nil
+  "If running in FSF Emacs, this variable stores the standard
+menu bar of VM internally.                  USR, 2011-02-27")
 
 (defun vm-menu-run-command (command &rest args)
   "Run COMMAND almost interactively, with ARGS.
@@ -726,6 +844,13 @@ set to the command name so that window configuration will be done."
 	     (not (vm-mime-plain-message-p (car vm-message-pointer)))))
     (error nil)))
 
+(defun vm-menu-can-convert-to-text/plain (layout)
+  (let ((type (car (vm-mm-layout-type layout))))
+    (or (equal (nth 1 (vm-mime-can-convert type)) "text/plain")
+	(and (equal type "message/external-body")
+	     (vm-menu-can-convert-to-text/plain
+	      (car (vm-mm-layout-parts layout)))))))
+
 (defun vm-menu-can-expunge-pop-messages-p ()
   (condition-case nil
       (save-excursion
@@ -763,7 +888,7 @@ set to the command name so that window configuration will be done."
 
 (defun vm-menu-create-subject-virtual-folder ()
   (interactive)
-  (vm-select-folder-buffer)
+  (vm-select-folder-buffer-and-validate 0 (vm-interactive-p))
   (setq this-command 'vm-create-virtual-folder)
   (vm-create-virtual-folder 'sortable-subject (regexp-quote
 	 			       (vm-so-sortable-subject
@@ -771,11 +896,18 @@ set to the command name so that window configuration will be done."
 
 (defun vm-menu-create-author-virtual-folder ()
   (interactive)
-  (vm-select-folder-buffer)
+  (vm-select-folder-buffer-and-validate 0 (vm-interactive-p))
   (setq this-command 'vm-create-virtual-folder)
   (vm-create-virtual-folder 'author (regexp-quote
 				     (vm-su-from (car vm-message-pointer)))))
 
+(defun vm-menu-mail-to ()
+  (interactive)
+  (vm-select-folder-buffer-and-validate 0 (vm-interactive-p))
+  (setq this-command 'vm-mail)
+  (vm-mail (vm-get-header-contents (car vm-message-pointer) "From:")))
+
+
 (defun vm-menu-xemacs-global-menubar ()
   (save-excursion
     (set-buffer (get-buffer-create "*scratch*"))
@@ -790,58 +922,60 @@ set to the command name so that window configuration will be done."
 	    (dummy (make-sparse-keymap)))
 	;; initialize all the vm-menu-fsfemacs-*-menu variables
 	;; with the menus.
-	(vm-easy-menu-define vm-menu-fsfemacs-help-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-help-menu (list dummy) nil
 			     vm-menu-help-menu)
-	(vm-easy-menu-define vm-menu-fsfemacs-dispose-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-dispose-menu (list dummy) nil
 			     (cons "Dispose" (nthcdr 4 vm-menu-dispose-menu)))
-	(vm-easy-menu-define vm-menu-fsfemacs-dispose-popup-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-dispose-popup-menu (list dummy) nil
 			     vm-menu-dispose-menu)
-;;	(vm-easy-menu-define vm-menu-fsfemacs-undo-menu (list dummy) nil
-;;			     (list "Undo" vm-menu-undo-menu))
-	(vm-easy-menu-define vm-menu-fsfemacs-virtual-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-undo-menu (list dummy) nil
+	  		     vm-menu-undo-menu)
+	(easy-menu-define vm-menu-fsfemacs-emacs-menu (list dummy) nil
+			     vm-menu-emacs-menu)
+	(easy-menu-define vm-menu-fsfemacs-virtual-menu (list dummy) nil
 			     vm-menu-virtual-menu)
-	(vm-easy-menu-define vm-menu-fsfemacs-sort-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-sort-menu (list dummy) nil
 			     vm-menu-sort-menu)
-	(vm-easy-menu-define vm-menu-fsfemacs-label-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-label-menu (list dummy) nil
 			     vm-menu-label-menu)
-	(vm-easy-menu-define vm-menu-fsfemacs-mark-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-mark-menu (list dummy) nil
 			     vm-menu-mark-menu)
-	(vm-easy-menu-define vm-menu-fsfemacs-send-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-send-menu (list dummy) nil
 			     vm-menu-send-menu)
-	(vm-easy-menu-define vm-menu-fsfemacs-motion-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-motion-menu (list dummy) nil
 			     vm-menu-motion-menu)
-;;	(vm-easy-menu-define vm-menu-fsfemacs-folders-menu (list dummy) nil
+;;	(easy-menu-define vm-menu-fsfemacs-folders-menu (list dummy) nil
 ;;			     vm-menu-folders-menu)
-	(vm-easy-menu-define vm-menu-fsfemacs-folder-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-folder-menu (list dummy) nil
 			     vm-menu-folder-menu)
-	(vm-easy-menu-define vm-menu-fsfemacs-vm-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-vm-menu (list dummy) nil
 			     vm-menu-vm-menu)
 	;; for mail mode
-	(vm-easy-menu-define vm-menu-fsfemacs-mail-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-mail-menu (list dummy) nil
 			     vm-menu-mail-menu)
 	;; subject menu
-	(vm-easy-menu-define vm-menu-fsfemacs-subject-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-subject-menu (list dummy) nil
 			     vm-menu-subject-menu)
 	;; author menu
-	(vm-easy-menu-define vm-menu-fsfemacs-author-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-author-menu (list dummy) nil
 			     vm-menu-author-menu)
 	;; url browser menu
-	(vm-easy-menu-define vm-menu-fsfemacs-url-browser-menu (list dummy) nil
+	(easy-menu-define vm-menu-fsfemacs-url-browser-menu (list dummy) nil
 			     vm-menu-url-browser-menu)
 	;; mailto url browser menu
-	(vm-easy-menu-define vm-menu-fsfemacs-mailto-url-browser-menu
+	(easy-menu-define vm-menu-fsfemacs-mailto-url-browser-menu
 			     (list dummy) nil
 			     vm-menu-url-browser-menu)
 	;; mime dispose menu
-	(vm-easy-menu-define vm-menu-fsfemacs-mime-dispose-menu
+	(easy-menu-define vm-menu-fsfemacs-mime-dispose-menu
 			     (list dummy) nil
 			     vm-menu-mime-dispose-menu)
 	;; attachment menu
-	(vm-easy-menu-define vm-menu-fsfemacs-attachment-menu
+	(easy-menu-define vm-menu-fsfemacs-attachment-menu
 			     (list dummy) nil
 			     vm-menu-attachment-menu)
 	;; image menu
-	(vm-easy-menu-define vm-menu-fsfemacs-image-menu
+	(easy-menu-define vm-menu-fsfemacs-image-menu
 			     (list dummy) nil
 			     vm-menu-image-menu)
 	;; block the global menubar entries in the map so that VM
@@ -866,7 +1000,7 @@ set to the command name so that window configuration will be done."
 		 (folder
 		  (cons "Folder" vm-menu-fsfemacs-folder-menu))
 		 (help
-		  (cons "Help!" vm-menu-fsfemacs-help-menu))
+		  (cons "Help" vm-menu-fsfemacs-help-menu))
 		 (label
 		  (cons "Label" vm-menu-fsfemacs-label-menu))
 		 (mark
@@ -880,27 +1014,33 @@ set to the command name so that window configuration will be done."
 		 (virtual
 		  (cons "Virtual" vm-menu-fsfemacs-virtual-menu))
 		 (emacs
-		  (cons "[Emacs]" 'vm-menu-toggle-menubar))
+		  (if (and (vm-menubar-buttons-possible-p) 
+			   vm-use-menubar-buttons)
+		      (cons "[Emacs Menubar]" 'vm-menu-toggle-menubar)
+		    (cons "Menubar" vm-menu-fsfemacs-emacs-menu)))
 		 (undo
-		  (cons "[Undo]" 'vm-undo))))
-	      cons
+		  (if (and (vm-menubar-buttons-possible-p)
+			   vm-use-menubar-buttons)
+		      (cons "[Undo]" 'vm-undo)
+		    (cons "Undo" vm-menu-fsfemacs-undo-menu)))))
+	      (cons nil)
 	      (vec (vector 'rootmenu 'vm nil))
 	      ;; menus appear in the opposite order that we
 	      ;; define-key them.
 	      (menu-list
 	       (if (consp vm-use-menus)
 		   (reverse vm-use-menus)
-		 (list 'help nil 'dispose 'virtual 'sort
-		       'label 'mark 'send 'motion 'folder))))
+		 (list 'help nil 'dispose 'undo 'virtual 'sort
+		       'label 'mark 'send 'motion 'folder)))
+	      (menu nil))
 	  (while menu-list
-	    (if (null (car menu-list))
+	    (setq menu (car menu-list))
+	    (if (null menu)
 		nil;; no flushright support in FSF Emacs
-	      (aset vec 2 (intern (concat "vm-menubar-"
-					  (symbol-name
-					   (car menu-list)))))
-	      (setq cons (assq (car menu-list) menu-alist))
+	      (aset vec 2 (intern (concat "vm-menubar-" (symbol-name menu))))
+	      (setq cons (assq menu menu-alist))
 	      (if cons
-		  (define-key map vec (eval (car (cdr cons))))))
+		  (define-key map vec (eval (cadr cons)))))
 	    (setq menu-list (cdr menu-list))))
 	(setq vm-mode-menu-map map)
 	(run-hooks 'vm-menu-setup-hook))))
@@ -917,7 +1057,7 @@ set to the command name so that window configuration will be done."
 	   (sort . vm-menu-sort-menu)
 	   (virtual . vm-menu-virtual-menu)
 	   (emacs . vm-menu-emacs-button)
-	   (undo . vm-menu-undo-menu)))
+	   (undo . vm-menu-xemacs-undo-button)))
 	cons
 	(menubar nil)
 	(menu-list vm-use-menus))
@@ -946,7 +1086,7 @@ set to the command name so that window configuration will be done."
   (interactive "e")
   ;; We should not need to do anything here for XEmacs.  The
   ;; default binding of mouse-3 is popup-mode-menu which does
-  ;; what we want for the normal case.  For special contexts,
+  ;; what we want for the normal case.  For special context,s
   ;; like when the mouse is over an URL, XEmacs has local keymap
   ;; support for extents.  Any context sensitive area should be
   ;; contained in an extent with a keymap that has mouse-3 bound
@@ -1079,15 +1219,35 @@ set to the command name so that window configuration will be done."
   (cond ((vm-menu-xemacs-menus-p)
 	 (set-menubar-dirty-flag))
 	((vm-menu-fsfemacs-menus-p)
-	 (force-mode-line-update))))
+	 ;; force-mode-line-update seems to have been buggy in Emacs
+	 ;; 21, 22, and 23.  So we do it ourselves.  USR, 2011-02-26
+	 ;; (force-mode-line-update t)
+	 (set-buffer-modified-p (buffer-modified-p))
+	 (when (and vm-user-interaction-buffer
+		    (buffer-live-p vm-user-interaction-buffer))
+	   (with-current-buffer vm-user-interaction-buffer
+	     (set-buffer-modified-p (buffer-modified-p)))))))
 
+(defun vm-menu-fsfemacs-add-vm-menu ()
+  "Add a menu or a menubar button to the Emacs menubar for switching
+to a VM menubar."
+  (if (and (vm-menubar-buttons-possible-p) vm-use-menubar-buttons)
+      (define-key vm-mode-map [menu-bar vm]
+	'(menu-item "[VM Menubar]" vm-menu-toggle-menubar))
+    (define-key vm-mode-map [menu-bar vm]
+      (cons "Menubar" (make-sparse-keymap "VM")))
+    (define-key vm-mode-map [menu-bar vm vm-toggle]
+      '(menu-item "Switch to VM Menubar" vm-menu-toggle-menubar))))
+      
 (defun vm-menu-toggle-menubar (&optional buffer)
+  "Toggle between the VM's dedicated menu bar and the standard Emacs
+menu bar.                                             USR, 2011-02-27"
   (interactive)
   (if buffer
       (set-buffer buffer)
-    (vm-select-folder-buffer))
+    (vm-select-folder-buffer-and-validate 0 (vm-interactive-p)))
   (cond ((vm-menu-xemacs-menus-p)
-	 (if (null (car (find-menu-item current-menubar '("XEmacs"))))
+	 (if (null (car (find-menu-item current-menubar '("[Emacs Menubar]"))))
 	     (set-buffer-menubar vm-menu-vm-menubar)
 	   ;; copy the current menubar in case it has been changed.
 	   (make-local-variable 'vm-menu-vm-menubar)
@@ -1096,7 +1256,7 @@ set to the command name so that window configuration will be done."
 	   (condition-case nil
 	       (add-menu-button nil vm-menu-vm-button nil)
 	     (void-function
-	      (add-menu-item nil "VM" 'vm-menu-toggle-menubar t))))
+	      (add-menu-item nil "Menubar" 'vm-menu-toggle-menubar t))))
 	 (vm-menu-set-menubar-dirty-flag)
 	 (vm-check-for-killed-summary)
 	 (and vm-summary-buffer
@@ -1112,12 +1272,12 @@ set to the command name so that window configuration will be done."
 	     (define-key vm-mode-map [menu-bar]
 	       (lookup-key vm-mode-menu-map [rootmenu vm]))
 	   (define-key vm-mode-map [menu-bar]
-	     (make-sparse-keymap))
-	   (define-key vm-mode-map [menu-bar vm]
-	     (cons "[VM]" 'vm-menu-toggle-menubar)))
+	     (make-sparse-keymap "Menu"))
+	   (vm-menu-fsfemacs-add-vm-menu))
 	 (vm-menu-set-menubar-dirty-flag))))
 
 (defun vm-menu-install-menubar ()
+  "Install the dedicated menu bar of VM.              USR, 2011-02-27"
   (cond ((vm-menu-xemacs-menus-p)
 	 (setq vm-menu-vm-menubar (vm-menu-make-xemacs-menubar))
 	 (set-buffer-menubar vm-menu-vm-menubar)
@@ -1131,6 +1291,8 @@ set to the command name so that window configuration will be done."
 	   (lookup-key vm-mode-menu-map [rootmenu vm])))))
 
 (defun vm-menu-install-menubar-item ()
+  "Install VM's menu on the current - presumably the standard - menu
+bar.						     USR, 2011-02-27"
   (cond ((and (vm-menu-xemacs-menus-p) (vm-menu-xemacs-global-menubar))
 	 (set-buffer-menubar (copy-sequence (vm-menu-xemacs-global-menubar)))
 	 (add-menu nil "VM" (cdr vm-menu-vm-menu)))
@@ -1142,6 +1304,7 @@ set to the command name so that window configuration will be done."
 	   (lookup-key vm-mode-menu-map [rootmenu])))))
 
 (defun vm-menu-install-vm-mode-menu ()
+  "This function strangely does nothing!               USR, 2011-02-27."
   ;; nothing to do here.
   ;; handled in vm-mouse.el
   (cond ((vm-menu-xemacs-menus-p)
@@ -1176,6 +1339,9 @@ set to the command name so that window configuration will be done."
 	       'vm-menu-popup-context-menu)))))
 
 (defun vm-menu-install-menus ()
+  "Install VM menus, either in the current menu bar or in a
+separate dedicated menu bar, depending on the value of
+`vm-use-menus'.                           USR, 2011-02-27"
   (cond ((consp vm-use-menus)
 	 (vm-menu-install-vm-mode-menu)
 	 (vm-menu-install-menubar)
@@ -1197,8 +1363,7 @@ set to the command name so that window configuration will be done."
       (setq menu (cons (vector "    "
 			       (list 'vm-menu-run-command
 				     ''vm-visit-virtual-folder (car folders))
-			       t
-			       (car folders))
+			       :suffix (car folders))
 		       menu)
 	    folders (cdr folders)))
     (and menu (setq menu (nreverse menu)
@@ -1210,7 +1375,7 @@ set to the command name so that window configuration will be done."
 	  (vm-menu-set-menubar-dirty-flag)
 	  (cond ((vm-menu-fsfemacs-menus-p)
 		 (makunbound 'vm-menu-fsfemacs-virtual-menu)
-		 (vm-easy-menu-define vm-menu-fsfemacs-virtual-menu
+		 (easy-menu-define vm-menu-fsfemacs-virtual-menu
 				      (list (make-sparse-keymap))
 				      nil
 				      vm-menu-virtual-menu)
@@ -1229,18 +1394,20 @@ set to the command name so that window configuration will be done."
       (setq menu (cons
 		  (vector "    "
 			  (cond
-			   ((and (stringp vm-recognize-pop-maildrops)
-				 (string-match vm-recognize-pop-maildrops
-					       (car folders))
+			   ((and (vm-pop-folder-spec-p (car folders))
 				 (setq foo (vm-pop-find-name-for-spec
 					    (car folders))))
 			    (list 'vm-menu-run-command
 				  ''vm-visit-pop-folder foo))
+			   ((and (vm-imap-folder-spec-p (car folders))
+				 (setq foo (vm-imap-folder-for-spec
+					    (car folders))))
+			    (list 'vm-menu-run-command
+				  'vm'visit-imap-folder foo))
 			   (t
 			    (list 'vm-menu-run-command
 				  ''vm-visit-folder (car folders))))
-			  t
-			  (car folders))
+			  :suffix (car folders))
 		       menu)
 	    folders (cdr folders)
 	    i (1+ i)))
@@ -1262,8 +1429,7 @@ set to the command name so that window configuration will be done."
 			(list (vector "    "
 				      (list 'vm-menu-run-command
 					    ''vm-visit-folder (car folders))
-				      t
-				      (car folders))))
+				      :suffix (car folders))))
 	    folders (cdr folders)))
     (setq tail (vm-member special vm-menu-folder-menu))
     (if (and menu tail)
@@ -1272,20 +1438,48 @@ set to the command name so that window configuration will be done."
 	  (vm-menu-set-menubar-dirty-flag)
 	  (cond ((vm-menu-fsfemacs-menus-p)
 		 (makunbound 'vm-menu-fsfemacs-folder-menu)
-		 (vm-easy-menu-define vm-menu-fsfemacs-folder-menu
+		 (easy-menu-define vm-menu-fsfemacs-folder-menu
 				      (list (make-sparse-keymap))
 				      nil
 				      vm-menu-folder-menu)
 		 (define-key vm-mode-menu-map [rootmenu vm vm-menubar-folder]
 		   (cons "Folder" vm-menu-fsfemacs-folder-menu))))))))
 
+(defun vm-customize ()
+  "Customize VM options."
+  (interactive)
+  (customize-group 'vm))
+
+(defun vm-view-news ()
+  "View NEWS for the current VM version."
+  (interactive)
+  (let* ((vm-dir (file-name-directory (locate-library "vm")))
+	 (doc-dirs (list (and vm-configure-docdir
+			       (expand-file-name vm-configure-docdir))
+			 (and vm-configure-datadir
+			       (expand-file-name vm-configure-datadir))
+			 (concat vm-dir "../")))
+	 doc-dir)
+    (while doc-dirs
+      (setq doc-dir (car doc-dirs))
+      (if (and doc-dir
+               (file-exists-p (expand-file-name "NEWS" doc-dir)))
+          (setq doc-dirs nil)
+	(setq doc-dirs (cdr doc-dirs))))
+    (view-file-other-frame (expand-file-name "NEWS" doc-dir))))
+
+(defun vm-view-manual ()
+  "View the VM manual."
+  (interactive)
+  (info "VM"))
+
 
 ;;; Muenkel Folders menu code
 
 (defvar vm-menu-hm-no-hidden-dirs t
   "*Hidden directories are suppressed in the folder menus, if non nil.")
 
-(defvar vm-menu-hm-hidden-file-list '("^\\..*" ".*\\.~[0-9]+~"))
+(defconst vm-menu-hm-hidden-file-list '("^\\..*" ".*\\.~[0-9]+~"))
 
 (defun vm-menu-hm-delete-folder (folder)
   "Query deletes a folder."
@@ -1296,11 +1490,11 @@ set to the command name so that window configuration will be done."
 	    (if (file-directory-p folder)
 		(delete-directory folder)
 	      (delete-file folder))
-	    (message "Folder deleted.")
+	    (vm-inform 5 "Folder deleted.")
 	    (vm-menu-hm-make-folder-menu)
 	    (vm-menu-hm-install-menu)
 	    )
-	(message "Aborted"))
+	(vm-inform 0 "Aborted"))
     (error "Folder %s does not exist." folder)
     (vm-menu-hm-make-folder-menu)
     (vm-menu-hm-install-menu)
@@ -1342,7 +1536,7 @@ set to the command name so that window configuration will be done."
 (defun vm-menu-hm-make-folder-menu ()
   "Makes a menu with the mail folders of the directory `vm-folder-directory'."
   (interactive)
-  (message "Building folders menu...")
+  (vm-inform 5 "Building folders menu...")
   (let ((folder-list (vm-menu-hm-tree-make-file-list vm-folder-directory))
 	(inbox-list (if (listp (car vm-spool-files))
 			(mapcar 'car vm-spool-files)
@@ -1399,7 +1593,7 @@ set to the command name so that window configuration will be done."
 		      "----"
 		      ["Rebuild Folders Menu" vm-menu-hm-make-folder-menu vm-folder-directory]
 		      ))))
-  (message "Building folders menu... done")
+  (vm-inform 5 "Building folders menu... done")
   (vm-menu-hm-install-menu))
 
 (defun vm-menu-hm-install-menu ()
@@ -1412,7 +1606,7 @@ set to the command name so that window configuration will be done."
 		(add-menu '("Folder") "Manipulate Folders"
 			  (cdr vm-menu-folders-menu) "Motion"))))
 	((vm-menu-fsfemacs-menus-p)
-	 (vm-easy-menu-define vm-menu-fsfemacs-folders-menu
+	 (easy-menu-define vm-menu-fsfemacs-folders-menu
 			      (list (make-sparse-keymap))
 			      nil
 			      vm-menu-folders-menu)
@@ -1422,7 +1616,7 @@ set to the command name so that window configuration will be done."
 
 ;;; Muenkel tree-menu code
 
-(defvar vm-menu-hm-tree-ls-flags "-aFLR"
+(defconst vm-menu-hm-tree-ls-flags "-aFLR"
   "*A String with the flags used in the function
 vm-menu-hm-tree-ls-in-temp-buffer for the ls command.
 Be careful if you want to change this variable.
@@ -1445,7 +1639,7 @@ The original flags are -aFLR.")
   (goto-char (point-min)))
 
 
-(defvar vm-menu-hm-tree-temp-buffername "*tree*"
+(defconst vm-menu-hm-tree-temp-buffername "*tree*"
   "Name of the temp buffers in tree.")
 
 
@@ -1575,6 +1769,4 @@ for the current directory (.) is inserted."
     )
   )
 
-(provide 'vm-menu)
-
 ;;; vm-menu.el ends here
